@@ -98,7 +98,11 @@ enum cram_encoding {
     E_SUBEXP             = 7,
     E_GOLOMB_RICE        = 8,
     E_GAMMA              = 9,
-    E_NUM_CODECS         = 10, /* Number of codecs, not a real one. */
+    E_XHUFFMAN           = 10, // To external block
+    E_XPACK              = 11, // Transform to sub-codec
+    E_XRLE               = 12, // Transform to sub-codec
+    E_XDELTA             = 13, // Transform to sub-codec
+    E_NUM_CODECS, /* Total number of codecs, not a real one. */
 };
 
 enum cram_external_type {
@@ -107,6 +111,8 @@ enum cram_external_type {
     E_BYTE               = 3,
     E_BYTE_ARRAY         = 4,
     E_BYTE_ARRAY_BLOCK   = 5,
+    E_SINT               = 6, // signed INT
+    E_SLONG              = 7, // signed LONG
 };
 
 /* External IDs used by this implementation (only assumed during writing) */
@@ -158,6 +164,7 @@ enum cram_DS_ID {
     DS_SC_len,
     DS_BB_len,
     DS_QQ_len,
+    DS_QS_len,
 
     DS_TC, // CRAM v1.0 tags
     DS_TM, // test
@@ -174,39 +181,76 @@ typedef struct {
     char    file_id[20];      // Filename or SHA1 checksum
 } cram_file_def;
 
+// NB, some of these are >= (eg >= 3).  This is so old code that checks
+// for Version 3 is also going to work on V4, V5 etc.
+// Therefore in if-elseif-else clauses, start with higher version
+// numbers first.
 #define CRAM_MAJOR_VERS(v) ((v) >> 8)
 #define CRAM_MINOR_VERS(v) ((v) & 0xff)
 #define IS_CRAM_1_VERS(fd) (CRAM_MAJOR_VERS((fd)->version)==1)
 #define IS_CRAM_2_VERS(fd) (CRAM_MAJOR_VERS((fd)->version)==2)
 #define IS_CRAM_3_VERS(fd) (CRAM_MAJOR_VERS((fd)->version)>=3)
+#define IS_CRAM_4_VERS(fd) (CRAM_MAJOR_VERS((fd)->version)>=4)
 
 struct cram_slice;
 
 enum cram_block_method {
+// Also see methmap[] in cram_compress_block() @ cram_io.c
     BM_ERROR  = -1,
     RAW    = 0,
-    GZIP   = 1,    // Z_FILTERED
+    GZIP   = 1, // Z_FILTERED
     BZIP2  = 2,
     LZMA   = 3,
-    RANS0  = 4,
-    BSC    = 5,
-    FQZ    = 6,
-    RANS1  = 10,   // Not externalised; stored as RANS (generic)
-    GZIP_RLE = 11, // Z_RLE, NB: not externalised in CRAM
-    GZIP_1 = 12,   // Z_DEFAULT_STRATEGY level 1, NB: not externalised in CRAM
+    RANS   = 4, RANS0     = RANS,
+    RANSPR = 5, RANS_PR0  = RANSPR,
+    ARITH  = 6, ARITH_PR0 = ARITH,
+    FQZ    = 7,
+    TOK3   = 8, NAME_TOK3 = TOK3,
 
-    RANS_PR0   = 13, // Parameterised r4x16pr rANS codecs.
-    RANS_PR1   = 14,
-    RANS_PR64  = 15,
-    RANS_PR65  = 16,
-    RANS_PR128 = 17,
-    RANS_PR129 = 18,
-    RANS_PR192 = 19,
-    RANS_PR193 = 20,
+    // Reserved for expansion
+    BSC    = 9,
+    ZSTD   = 10,
 
-    NAME_TOK3  = 21,
+    // Methods not externalised, but used in metrics.
+    // Externally they become one of the above methods.
+    FQZ_b, FQZ_c, FQZ_d, // Various preset FQZ methods
 
-    BLOCK_METHOD_END = 31
+  //RANS0,       // Order 0
+    RANS1,       // Order 1
+
+    GZIP_RLE,    // Z_RLE
+    GZIP_1,      // Z_DEFAULT_STRATEGY level 1, NB: not externalised in CRAM
+
+  //RANS_PR0,    // Order 0
+    RANS_PR1,    // Order 1
+    RANS_PR64,   // O0 + RLE
+    RANS_PR9,    // O1 + X4
+    RANS_PR128,  // O0 + Pack
+    RANS_PR129,  // O1 + Pack
+    RANS_PR192,  // O0 + RLE + pack
+    RANS_PR193,  // O1 + RLE + pack
+
+  //NAME_TOK3,   // tok+rans
+    NAME_TOKA,   // tok+arith
+
+  //ARITH_PR0,   // Order 0
+    ARITH_PR1,   // Order 1
+    ARITH_PR64,  // O0 + RLE
+    ARITH_PR9,   // O1 + X4
+    ARITH_PR128, // O0 + Pack
+    ARITH_PR129, // O1 + Pack
+    ARITH_PR192, // O0 + RLE + pack
+    ARITH_PR193, // O1 + RLE + pack
+    // End with ARITH_PR193 == 31
+
+    // NB: must end on no more than 31 unless we change to a
+    // 64-bit method type.
+
+    ZSTD_1,      // ZSTD level 1 is sometimes enough, and better!
+
+    // For purposes of returning a more precise orig_method field
+    RANS_PR65, RANS_PR32, RANS_PR8,
+    ARITH_PR65, ARITH_PR32, ARITH_PR8,
 };
 
 enum cram_content_type {
@@ -220,7 +264,7 @@ enum cram_content_type {
 };
 
 /* Maximum simultaneous codecs allowed, 1 per bit */
-#define CRAM_MAX_METHOD 32
+#define CRAM_MAX_METHOD 33
 
 /* Compression metrics */
 typedef struct {
@@ -233,12 +277,12 @@ typedef struct {
     int sz[CRAM_MAX_METHOD];
 
     // resultant method from trials
-    int method;
+    int64_t method;
     int strat;
 
     // Revisions of method, to allow culling of continually failing ones.
     int cnt[CRAM_MAX_METHOD];
-    int revised_method;
+    int64_t revised_method;
 
     double extra[CRAM_MAX_METHOD];
 
@@ -284,15 +328,12 @@ typedef struct {
     int32_t *landmark;
 
     /* Flags from preservation map */
-    int mapped_qs_included;
-    int unmapped_qs_included;
-    int unmapped_placed;
-    int qs_included;
     int read_names_included;
     int AP_delta;
+    int no_ref;
+    int qs_seq_orient; // 1 => same as seq. 0 => original orientation
     // indexed by ref-base and subst. code
     char substitution_matrix[5][4];
-    int no_ref;
 
     // TD Dictionary as a concatenated block
     cram_block *TD_blk;  // Tag Dictionary
@@ -308,6 +349,9 @@ typedef struct {
 
     char *uncomp; // A single block of uncompressed data
     size_t uncomp_size, uncomp_alloc;
+
+    // Total codec count, used for index to block_by_id for transforms
+    int ncodecs;
 } cram_block_compression_hdr;
 
 typedef struct cram_map {
@@ -322,6 +366,7 @@ typedef struct cram_map {
 typedef struct {
     struct cram_codec *codec;
     cram_block *blk;
+    cram_block *blk2;
     cram_metrics *m;
 } cram_tag_map;
 
@@ -387,6 +432,7 @@ typedef struct {
     int last_slice;              // number of reads in last slice (0 for 1st)
     int multi_seq;               // true if packing multi seqs per cont/slice
     int unsorted;		 // true is AP_delta is 0.
+    int qs_seq_orient;           // 1 => same as seq. 0 => original orientation
 
     /* Copied from fd before encoding, to allow multi-threading */
     int64_t ref_start, first_base, last_base, ref_id, ref_end;
@@ -409,6 +455,8 @@ typedef struct {
     uint32_t crc32;       // Raw container bytes CRC
 
     uint64_t s_num_bases; // number of bases in this slice
+
+    uint32_t n_mapped;    // Number of mapped reads
 } cram_container;
 
 /*
@@ -429,6 +477,7 @@ typedef struct {
     int32_t mate_ref_id;
     int64_t mate_pos;     // NP
     int64_t tlen;         // TS
+    int64_t explicit_tlen;// TS, but PNEXT/RNEXT still need auto-computing
 
     // Auxiliary data
     int32_t ntags;        // TC
@@ -586,6 +635,7 @@ typedef struct cram_slice {
     int ref_start;           // start position of current reference;
     int ref_end;             // end position of current reference;
     int ref_id;
+    char *cons;              // from ref_start to ref_end inclusive
 
     uint32_t BD_crc;         // base call digest
     uint32_t SD_crc;         // quality score digest
@@ -744,7 +794,39 @@ typedef struct {
 } cram_fd_output_buffer;
 #endif
 
-typedef struct {
+struct cram_fd;
+typedef struct varint_vec {
+    // Returns number of bytes decoded from fd, 0 on error
+    int (*varint_decode32_crc)(struct cram_fd *fd, int32_t *val_p, uint32_t *crc);
+    int (*varint_decode64_crc)(struct cram_fd *fd, int64_t *val_p, uint32_t *crc);
+
+    //int (*varint_get32)(const char *cp, const char *endp, int32_t *val_p);
+    //int (*varint_get64)(const char *cp, const char *endp, int64_t *val_p);
+
+    // Returns the value and increments *cp.  Sets err to 1 iff an error occurs.
+    // NOTE: Does not set err to 0 on success.
+    int64_t (*varint_get32) (char **cp, const char *endp, int *err);
+    int64_t (*varint_get32s)(char **cp, const char *endp, int *err);
+    int64_t (*varint_get64) (char **cp, const char *endp, int *err);
+    int64_t (*varint_get64s)(char **cp, const char *endp, int *err);
+
+    // Returns the number of bytes written, <= 0 on error.
+    int (*varint_put32) (char *cp, const char *endp, int32_t val_p);
+    int (*varint_put32s)(char *cp, const char *endp, int32_t val_p);
+    int (*varint_put64) (char *cp, const char *endp, int64_t val_p);
+    int (*varint_put64s)(char *cp, const char *endp, int64_t val_p);
+
+    // Returns the number of bytes written, <= 0 on error.
+    int (*varint_put32_blk) (cram_block *blk, int32_t val_p);
+    int (*varint_put32s_blk)(cram_block *blk, int32_t val_p);
+    int (*varint_put64_blk) (cram_block *blk, int64_t val_p);
+    int (*varint_put64s_blk)(cram_block *blk, int64_t val_p);
+
+    // Returns number of bytes needed to encode 'val'
+    int (*varint_size)(int64_t val);
+} varint_vec;
+
+typedef struct cram_fd {
     FILE                 *fp_in;
 #if defined(CRAM_IO_CUSTOM_BUFFERING)
     cram_fd_input_buffer            *fp_in_buffer;
@@ -801,13 +883,17 @@ typedef struct {
     int bases_per_slice;
     int slices_per_container;
     int embed_ref;
+    int embed_cons;
     int no_ref;
     int ignore_md5;
     int use_bz2;
     int use_rans;
     int use_lzma;
     int use_bsc;
+    int use_zstd;
     int use_fqz;
+    int use_tok;
+    int use_arith;
     int shared_ref;
     enum quality_binning binning;
     unsigned int required_fields;
@@ -825,8 +911,10 @@ typedef struct {
     off_t first_container;
     int eof;
     int last_slice;                     // number of recs encoded in last slice
-    int multi_seq;
+    int last_RI;                        // number of references encoded in last container
+    int multi_seq, multi_seq_user;
     int unsorted;
+    int last_mapped;                    // number of mapped reads in last container
     int empty_container; 		// Marker for EOF block
     
     // thread pool
@@ -844,6 +932,10 @@ typedef struct {
     int lossy_read_names;
     int preserve_aux_order;             // if set implies emitting RG, MD and NM
     int preserve_aux_size;              // does not replace 'i' with 'c' etc in aux.
+
+    // variable integer decoding callbacks.
+    // This changed in CRAM4.0 to a data-size agnostic encoding.
+    varint_vec vv;
 } cram_fd;
 
 #if defined(CRAM_IO_CUSTOM_BUFFERING)
@@ -971,7 +1063,10 @@ enum cram_option {
     CRAM_OPT_WITH_BGZIP_INDEX,
     CRAM_OPT_OUTPUT_BGZIP_IDX,
     CRAM_OPT_USE_BSC,
+    CRAM_OPT_USE_ZSTD,
     CRAM_OPT_USE_FQZ,
+    CRAM_OPT_EMBED_CONS,
+    CRAM_OPT_USE_TOK,
 };
 
 /* BF bitfields */
@@ -1005,7 +1100,9 @@ enum cram_option {
 #define CRAM_FLAG_DETACHED             (1<<1)
 #define CRAM_FLAG_MATE_DOWNSTREAM      (1<<2)
 #define CRAM_FLAG_NO_SEQ               (1<<3)
-#define CRAM_FLAG_MASK                 ((1<<4)-1)
+#define CRAM_FLAG_EXPLICIT_TLEN        (1<<4)
+#define CRAM_FLAG_TLEN0                (1<<5)
+#define CRAM_FLAG_MASK                 ((1<<6)-1)
 
 /* Internal only */
 #define CRAM_FLAG_STATS_ADDED  (1<<30)
