@@ -157,7 +157,26 @@ void dump_name_block(cram_block *b, int verbose) {
 }
 
 void dump_tag_block(cram_block *b, int verbose) {
-    return dump_core_block(b, verbose);
+    dump_core_block(b, verbose);
+}
+
+static int cram_decode_i64(cram_fd *fd, cram_container *c, cram_slice *s,
+			   cram_block *blk, int ds, int64_t *tlen) {
+    int out_sz = 1, r = 0;
+
+    if (!c->comp_hdr->codecs[ds]) return -1;
+    if (CRAM_MAJOR_VERS(fd->version) < 4) {
+        int32_t i32;
+        r |= c->comp_hdr->codecs[ds]
+            ->decode(s, c->comp_hdr->codecs[ds], blk,
+                     (char *)&i32, &out_sz);
+        *tlen = i32;
+    } else {
+        r |= c->comp_hdr->codecs[ds]
+            ->decode(s, c->comp_hdr->codecs[ds], blk,
+                     (char *)tlen, &out_sz);
+    }
+    return r;
 }
 
 int main(int argc, char **argv) {
@@ -363,6 +382,15 @@ int main(int argc, char **argv) {
 		    bmax = s->block[id]->content_id;
 	    }
 
+	    // Temporary copy of the actual method so we
+	    // can distinguish between CRAM format (RANS) and
+	    // "orig_method" format (RANS0 or RANS1).
+	    uint32_t *bmethod = malloc(s->hdr->num_blocks * sizeof(*bmethod));
+	    if (!bmethod)
+		return 1;
+	    for (id = 0; id < s->hdr->num_blocks; id++)
+		bmethod[id] = s->block[id]->method;
+
 	    for (id = 0; id < s->hdr->num_blocks; id++)
 		cram_uncompress_block(s->block[id]);
 
@@ -370,6 +398,7 @@ int main(int argc, char **argv) {
 	    if (verbose) {
 		cram_block *b = s->block[0];
 		int32_t i32, bf, fn, prev_pos, rl;
+		int64_t i64;
 		unsigned char cf;
 		int out_sz, r, f;
 		int rec;
@@ -405,8 +434,16 @@ int main(int argc, char **argv) {
 		    r = c->comp_hdr->codecs[DS_RL]->decode(s,c->comp_hdr->codecs[DS_RL], b, (char *)&rl, &out_sz);
 		    printf("RL = %d (ret %d, out_sz %d)\n", rl, r, out_sz);
 
-		    r = c->comp_hdr->codecs[DS_AP]->decode(s,c->comp_hdr->codecs[DS_AP], b, (char *)&i32, &out_sz);
-		    printf("AP = %d (ret %d, out_sz %d)\n", i32, r, out_sz);
+		    int64_t pos;
+		    if (CRAM_MAJOR_VERS(fd->version) < 4) {
+			r = c->comp_hdr->codecs[DS_AP]->decode(s,c->comp_hdr->codecs[DS_AP],
+							       b, (char *)&i32, &out_sz);
+			pos = i32;
+		    } else {
+			r = c->comp_hdr->codecs[DS_AP]->decode(s,c->comp_hdr->codecs[DS_AP],
+							       b, (char *)&pos, &out_sz);
+		    }
+		    printf("AP = %"PRId64" (ret %d, out_sz %d)\n", pos, r, out_sz);
 
 		    r = c->comp_hdr->codecs[DS_RG]->decode(s,c->comp_hdr->codecs[DS_RG], b, (char *)&i32, &out_sz);
 		    printf("RG = %d (ret %d, out_sz %d)\n", i32, r, out_sz);
@@ -443,15 +480,21 @@ int main(int argc, char **argv) {
 			r = c->comp_hdr->codecs[DS_NS]->decode(s,c->comp_hdr->codecs[DS_NS], b, (char *)&i32, &out_sz);
 			printf("NS = %d (ret %d, out_sz %d)\n", i32, r, out_sz);
 
-			r = c->comp_hdr->codecs[DS_NP]->decode(s,c->comp_hdr->codecs[DS_NP], b, (char *)&i32, &out_sz);
-			printf("NP = %d (ret %d, out_sz %d)\n", i32, r, out_sz);
+			r = cram_decode_i64(fd, c, s, b, DS_NP, &i64);
+			printf("NP = %"PRId64" (ret %d, out_sz %d)\n", i64, r, out_sz);
 
-			r = c->comp_hdr->codecs[DS_TS]->decode(s,c->comp_hdr->codecs[DS_TS], b, (char *)&i32, &out_sz);
-			printf("TS = %d (ret %d, out_sz %d)\n", i32, r, out_sz);
-		    } else if (cf & CRAM_FLAG_MATE_DOWNSTREAM) {
-			puts("Not detached, and mate is downstream");
-			r = c->comp_hdr->codecs[DS_NF]->decode(s,c->comp_hdr->codecs[DS_NF], b, (char *)&i32, &out_sz);
-			printf("NF = %d+%d = %d (ret %d, out_sz %d)\n", i32, rec+1, i32+rec+1, r, out_sz);
+			r = cram_decode_i64(fd, c, s, b, DS_TS, &i64);
+			printf("TS = %"PRId64" (ret %d, out_sz %d)\n", i64, r, out_sz);
+		    } else {
+			if (cf & CRAM_FLAG_MATE_DOWNSTREAM) {
+			    puts("Not detached, and mate is downstream");
+			    r = c->comp_hdr->codecs[DS_NF]->decode(s,c->comp_hdr->codecs[DS_NF], b, (char *)&i32, &out_sz);
+			    printf("NF = %d+%d = %d (ret %d, out_sz %d)\n", i32, rec+1, i32+rec+1, r, out_sz);
+			}
+			if (cf & CRAM_FLAG_EXPLICIT_TLEN) {
+			    r = cram_decode_i64(fd, c, s, b, DS_TS, &i64);
+			    printf("TS = %"PRId64" (ret %d, out_sz %d, explicit)\n", i64, r, out_sz);
+			}
 		    }
 
 		    if (IS_CRAM_1_VERS(fd)) {
@@ -471,20 +514,21 @@ int main(int argc, char **argv) {
 			    key[2] = id&0xff;
 			    printf("%3d: TN= %.3s\n", f, key);
 
-			    printf("id=%d\n", id);
-			    if ((m = map_find(c->comp_hdr->tag_encoding_map,
-					      (unsigned char *)key, id))) {
-				int i, out_sz;
+			    if (key[2] != '*') {
+				if ((m = map_find(c->comp_hdr->tag_encoding_map,
+						  (unsigned char *)key, id))) {
+				    int i, out_sz;
 
-				BLOCK_SIZE(tag) = 0;
-				r = m->codec->decode(s, m->codec, b, (char *)tag, &out_sz);
-				printf("%3d: Val", f);
-				for(i = 0; i < out_sz; i++) {
-				    printf(" %02x", BLOCK_DATA(tag)[i]);
+				    BLOCK_SIZE(tag) = 0;
+				    r = m->codec->decode(s, m->codec, b, (char *)tag, &out_sz);
+				    printf("%3d: Val", f);
+				    for(i = 0; i < out_sz; i++) {
+					printf(" %02x", BLOCK_DATA(tag)[i]);
+				    }
+				    printf("\n");
+				} else {
+				    fprintf(stderr, "*** ERROR: unrecognised aux key ***\n");
 				}
-				printf("\n");
-			    } else {
-				fprintf(stderr, "*** ERROR: unrecognised aux key ***\n");
 			    }
 
 			    cram_free_block(tag);
@@ -512,20 +556,21 @@ int main(int argc, char **argv) {
 			    id = (key[0]<<16) | (key[1]<<8) | key[2];
 			    printf("%3d: TN= %.3s\n", f, key);
 
-			    printf("id=%d\n", id);
-			    if ((m = map_find(c->comp_hdr->tag_encoding_map,
-					      (unsigned char *)key, id))) {
-				int i, out_sz;
+			    if (key[2] != '*') {
+				if ((m = map_find(c->comp_hdr->tag_encoding_map,
+						  (unsigned char *)key, id))) {
+				    int i, out_sz;
 
-				BLOCK_SIZE(tag) = 0;
-				r = m->codec->decode(s, m->codec, b, (char *)tag, &out_sz);
-				printf("%3d: Val", f);
-				for(i = 0; i < out_sz; i++) {
-				    printf(" %02x", BLOCK_DATA(tag)[i]);
+				    BLOCK_SIZE(tag) = 0;
+				    r = m->codec->decode(s, m->codec, b, (char *)tag, &out_sz);
+				    printf("%3d: Val", f);
+				    for(i = 0; i < out_sz; i++) {
+					printf(" %02x", BLOCK_DATA(tag)[i]);
+				    }
+				    printf("\n");
+				} else {
+				    fprintf(stderr, "*** ERROR: unrecognised aux key ***\n");
 				}
-				printf("\n");
-			    } else {
-				fprintf(stderr, "*** ERROR: unrecognised aux key ***\n");
 			    }
 
 			    cram_free_block(tag);
@@ -724,8 +769,8 @@ int main(int argc, char **argv) {
 		printf("\n\tBlock %d/%d\n", id+1, s->hdr->num_blocks);
 		printf("\t    Size:         %d comp / %d uncomp\n",
 		       b->comp_size, b->uncomp_size);
-		printf("\t    Method:       %s\n",
-		       cram_block_method2str(b->orig_method));
+		printf("\t    Method:       %s\t(%d)\n",
+		       cram_block_method2str(b->orig_method), bmethod[id]);
 		struct {
 		    int id;
 		    enum cram_block_method method;
@@ -798,6 +843,7 @@ int main(int argc, char **argv) {
 		}
 	    }
 
+	    free(bmethod);
 	    cram_free_slice(s);
 	}
 
@@ -859,7 +905,7 @@ int main(int argc, char **argv) {
 	    putchar((HashTableSearch(dc_h, (char *)&id_type, sizeof(id_type)))?'1':' ');
 	    id_type.method = RANS_PR64;
 	    putchar((HashTableSearch(dc_h, (char *)&id_type, sizeof(id_type)))?'4':' ');
-	    id_type.method = RANS_PR65;
+	    id_type.method = RANS_PR9;
 	    putchar((HashTableSearch(dc_h, (char *)&id_type, sizeof(id_type)))?'5':' ');
 	    id_type.method = RANS_PR128;
 	    putchar((HashTableSearch(dc_h, (char *)&id_type, sizeof(id_type)))?'8':' ');
