@@ -39,9 +39,6 @@
 #include "io_lib_config.h"
 #endif
 
-// Enable if we want V3.1 support.  TODO: add a configure param for this
-#define HAVE_FQZ
-
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -92,7 +89,7 @@ static char *detect_format(char *fn) {
 
 static void usage(FILE *fp) {
     fprintf(fp, "  -=- sCRAMble -=-     version %s\n", PACKAGE_VERSION);
-    fprintf(fp, "Author: James Bonfield, Wellcome Trust Sanger Institute. 2013-2018\n\n");
+    fprintf(fp, "Author: James Bonfield, Wellcome Trust Sanger Institute. 2013-2020\n\n");
 
     fprintf(fp, "Usage:    scramble [options] [input_file [output_file]]\n");
 
@@ -116,6 +113,7 @@ static void usage(FILE *fp) {
     fprintf(fp, "    -x             [Cram] Non-reference based encoding.\n");
     fprintf(fp, "    -M             [Cram] Use multiple references per slice.\n");
     fprintf(fp, "    -m             [Cram] Generate MD and NM tags.\n");
+    fprintf(fp, "    -a             [Cram] Also compress using arithmetic coder (V3.1+).\n");
 #ifdef HAVE_LIBBZ2
     fprintf(fp, "    -j             [Cram] Also compress using bzip2.\n");
 #endif
@@ -124,8 +122,12 @@ static void usage(FILE *fp) {
 #endif
 #ifdef HAVE_LIBBSC
     fprintf(fp, "    -J             [Cram] Also compression using libbsc (V3.1+)\n");
+#endif
+#ifdef HAVE_ZSTD
+    fprintf(fp, "    -z             [Cram] Also compression using zstd (V3.1+)\n");
 #endif
     fprintf(fp, "    -f             [Cram] Also compression using fqzcomp (V3.1+)\n");
+    fprintf(fp, "    -T             [Cram] Also compression using name tokeniser (V3.1+)\n");
     fprintf(fp, "    -n             [Cram] Discard read names where possible.\n");
     fprintf(fp, "    -P             Preserve all aux tags (incl RG,NM,MD)\n");
     fprintf(fp, "    -p             Preserve aux tag sizes ('i', 's', 'c')\n");
@@ -136,6 +138,7 @@ static void usage(FILE *fp) {
     fprintf(fp, "    -!             Disable all checking of checksums\n");
     fprintf(fp, "    -g FILE        Convert to Bam using index (file.gzi)\n");
     fprintf(fp, "    -G FILE        Output Bam index when bam input(file.gzi)\n");
+    fprintf(fp, "    -X mode        [Cram] Mode is fast, normal, small or archive.\n");
 }
 
 int main(int argc, char **argv) {
@@ -144,10 +147,11 @@ int main(int argc, char **argv) {
     char imode[10], *in_f = "", omode[10], *out_f = "", *index_fn = NULL, *index_out_fn = NULL;
     int level = '\0'; // nul terminate string => auto level
     int c, verbose = 0;
-    int s_opt = 0, S_opt = 0, embed_ref = 0, ignore_md5 = 0, decode_md = 0;
+    int s_opt = 0, S_opt = 0, embed_ref = 0, embed_cons = 0, ignore_md5 = 0, decode_md = 0;
     char *ref_fn = NULL;
     int start, end, multi_seq = -1, no_ref = 0;
-    int use_bz2 = 0, use_bsc = 0, use_lzma = 0, use_fqz = 0;
+    int use_bz2 = 0, use_bsc = 0, use_lzma = 0, use_fqz = 0, use_tok = 0, use_arith = 0, use_zstd = 0;
+    double vers = 3.0; // 3.0, 3.1, 4.0, etc
     char ref_name[1024] = {0};
     refs_t *refs;
     int nthreads = 1;
@@ -160,14 +164,46 @@ int main(int argc, char **argv) {
     int bases_per_slice = 0;
     int lossy_read_names = 0;
     int preserve_aux_order = 0;
-    int preserve_aux_size = 0; 
-    int add_pg = 1;   
+    int preserve_aux_size = 0;
+    int add_pg = 1;
+    int archive = 0;
 
     scram_init();
 
     /* Parse command line arguments */
-    while ((c = getopt(argc, argv, "u0123456789hvs:S:V:r:xXeI:O:R:!MmjJZt:BN:F:Hb:nPpqg:G:f")) != -1) {
+    while ((c = getopt(argc, argv, "u0123456789hvs:S:V:r:xeEI:O:R:!MmajJzZt:BN:F:Hb:nPpqg:G:fTX:")) != -1) {
 	switch (c) {
+	case 'X':
+	    if (strcmp(optarg, "default") == 0 || strcmp(optarg, "normal") == 0) {
+		// nothing for 3.0
+		if (vers >= 3.099)
+		    use_tok = 1;
+	    } else if (strcmp(optarg, "fast") == 0) {
+		if (!level) level = '1';
+		s_opt = 1000;
+		use_tok = 0;
+	    } else if (strcmp(optarg, "small") == 0) {
+		if (vers >= 3.099)
+		    use_fqz = use_tok = 1;
+		else
+		    use_bz2 = 1;
+		if (s_opt != 10000) s_opt = 25000;
+	    } else if (strcmp(optarg, "archive") == 0) {
+		archive = 1;
+		use_bz2 = 1;
+		if (vers >= 3.099)
+		    use_arith = use_fqz = use_tok = 1;
+		if (level >= '7')
+		    use_lzma = 1;
+		if (s_opt != 10000) s_opt = 100000;
+	    } else {
+		fprintf(stderr, "Unknown parameter set: choose 'fast', 'normal/default', "
+			"small or 'archive'\n");
+		fprintf(stderr, "Assuming default\n");
+	    }
+	    bases_per_slice = s_opt * 500; // guesswork...
+	    break;
+
 	case 'F':
 	    sam_fields = strtol(optarg, NULL, 0); // undocumented for testing
 	    break;
@@ -175,6 +211,8 @@ int main(int argc, char **argv) {
 	case '0': case '1': case '2': case '3': case '4':
 	case '5': case '6': case '7': case '8': case '9':
 	    level = c;
+	    if (archive && level >= 7)
+		use_lzma = 1;
 	    break;
 	    
 	case 'u':
@@ -211,19 +249,23 @@ int main(int argc, char **argv) {
 	    break;
 
 	case 'V':
+	    vers = atof(optarg);
 	    if (cram_set_option(NULL, CRAM_OPT_VERSION, optarg))
 		return 1;
+	    if (vers >= 3.099)
+		use_tok = 1;
 	    break;
 
 	case 'r':
 	    ref_fn = optarg;
 	    break;
 
-	case 'X':
-	    fprintf(stderr, "-X is deprecated in favour of -e.\n");
 	case 'e':
 	    embed_ref = 1;
 	    break;
+	case 'E':
+	    embed_cons = 1;
+	    break;
 
 	case 'x':
 	    no_ref = 1;
@@ -271,6 +313,10 @@ int main(int argc, char **argv) {
 	    multi_seq = 1;
 	    break;
 
+	case 'a':
+	    use_arith = 1;
+	    break;
+
 	case 'j':
 #ifdef HAVE_LIBBZ2
 	    use_bz2 = 1;
@@ -289,6 +335,15 @@ int main(int argc, char **argv) {
 		    " version.\nPlease recompile.\n");
 #endif
 
+#ifdef HAVE_ZSTD
+	case 'z':
+	    use_zstd = 1;
+	    break;
+#else
+	    fprintf(stderr, "Warning: zstd support is not compiled into this"
+		    " version.\nPlease recompile.\n");
+#endif
+
 	case 'Z':
 #ifdef HAVE_LIBLZMA
 	    use_lzma = 1;
@@ -299,12 +354,11 @@ int main(int argc, char **argv) {
 	    break;
 
 	case 'f':
-#ifdef HAVE_FQZ
 	    use_fqz = 1;
-#else
-	    fprintf(stderr, "Warning: FQZ support is not compiled into this"
-		    " version.\nPlease recompile.\n");
-#endif
+	    break;
+
+	case 'T':
+	    use_tok = 1;
 	    break;
 
 	case 't':
@@ -350,8 +404,8 @@ int main(int argc, char **argv) {
 	}
     }    
 
-    if (cram_default_version() <= 300 && (use_bsc || use_fqz)) {
-	fprintf(stderr, "Libbsc and/or fqzcomp codecs are only permitted in CRAM v3.1 and 4.0.\n"
+    if (cram_default_version() <= 300 && (use_bsc || use_fqz || use_zstd)) {
+	fprintf(stderr, "Libbsc, ZSTD and/or fqzcomp codecs are only permitted in CRAM v3.1 and 4.0.\n"
 		"Note these CRAM versions are a technology demonstration only.\n"
 		"Future versions of Scramble may not be able to read these files.\n");
 	return 1;
@@ -444,6 +498,17 @@ int main(int argc, char **argv) {
 	}
     }
 
+    if (embed_cons) {
+	if (scram_get_header(in)->sort_order == ORDER_NAME ||
+	    scram_get_header(in)->sort_order == ORDER_UNSORTED) {
+	    fprintf(stderr, "Embeded consensus with non-coordinate sorted data is "
+		    "not supported.\n");
+	} else {
+	    if (scram_set_option(out, CRAM_OPT_EMBED_CONS, embed_cons))
+		return 1;
+	}
+    }
+
     if (use_bz2)
 	if (scram_set_option(out, CRAM_OPT_USE_BZIP2, use_bz2))
 	    return 1;
@@ -452,6 +517,10 @@ int main(int argc, char **argv) {
 	if (scram_set_option(out, CRAM_OPT_USE_BSC, use_bsc))
 	    return 1;
 
+    if (use_zstd)
+	if (scram_set_option(out, CRAM_OPT_USE_ZSTD, use_zstd))
+	    return 1;
+
     if (use_lzma)
 	if (scram_set_option(out, CRAM_OPT_USE_LZMA, use_lzma))
 	    return 1;
@@ -460,6 +529,14 @@ int main(int argc, char **argv) {
 	if (scram_set_option(out, CRAM_OPT_USE_FQZ, use_fqz))
 	    return 1;
 
+    if (use_tok)
+	if (scram_set_option(out, CRAM_OPT_USE_TOK, use_tok))
+	    return 1;
+
+    if (use_arith)
+	if (scram_set_option(out, CRAM_OPT_USE_ARITH, use_arith))
+	    return 1;
+
     if (binning != BINNING_NONE)
 	if (scram_set_option(out, CRAM_OPT_BINNING, binning))
 	    return 1;
